
(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *		use is subject to a specific license, obtainable from LSV.
 *)

open Error
open Control
open Cparse
open Zones
open Glue
open Glue_log
open Sigs
open Api
open Prglue

type abs_info = (loc_ppoint option, (loc_ppoint, abs_env) Hashtbl.t) Hashtbl.t
    (* envoie chaque point de programme vers un environnement abstrait.
     Si le point de programme pp n'est pas dans l'abs_info, ceci
     signifie qu'il est envoye vers l'environnement bottom (code mort). *)

type sig_info = (loc_ppoint option, (loc_instr, sig_set) Hashtbl.t) Hashtbl.t
    (* envoie chaque instruction (avec eventuellement un point de programme
     identifiant la ligne appelant la fonction courante) vers la liste
     des signaux pouvant etre lances par cette instruction.
     Si l'instruction i est envoyee vers une table de hash H, alors H est
     non vide.  De plus H envoie chaque contexte ctx possible vers une liste
     de signaux non vides.  Autrement dit, les cas ou les listes de signaux
     sont vides ne sont pas enregistres. *)

type function_info = (string, (loc_ppoint option,
			       abs_value AbsEnv.t option ref *
			       abs_value option ref *
			       abs_addr ref *
			       sig_set ref) Hashtbl.t) Hashtbl.t
    (* envoie chaque nom de fonction et chaque contexte d'appel
     vers:
     - un(e moitie d') environnement representant les valeurs des variables
     globales a la fin de la fonction, ou None si la fonction ne retourne pas;
     - une valeur abstraite representant les valeurs de retour possibles,
     ou None si la fonction ne retourne pas ou bien est de type de retour void;
     - une adresse abstraite representant tous les effets de bord
     que cette fonction effectue possiblement lorsqu'elle est appelee
     dans ce contexte;
     - un ensemble de signaux que l'appel a la fonction peut lancer.
     *)

type functions_not_found = (string, unit) Hashtbl.t
    (* ensemble de toutes les fonctions dont on ne connait pas le source
     et qui ont ete appelees par au moins une fonction dans le code analyse. *)

(* Pour les elargissements, noter que les "node" des graphes de controle
 sont des entiers.  En fait, on a deja effectue le calcul du tri topologique
 inverse, et il suffit pour decider si on doit faire ou non un elargissement
 lors du franchissement de la transition m -> n de regarder si m>=n (elargissement)
 ou m<n (pas d'elargissement).
 Mais noter que ca ne suffit pas a traiter des elargissements en general,
 contrairement au cours.  Ca ne sert qu'aux elargissements *intraproceduraux*
 (a l'interieur d'une meme procedure).  Pour les elargissements *interproceduraux*
 (appels recursifs), on maintient un ensemble de points de programmes
 correspondant a des sites d'appels au-dessus de nous dans la pile. *)

module CallContext = Set.Make (struct
				   type t = ppoint
				   let compare = Pervasives.compare
			       end)
(*
  sig
    type elt = Control.ppoint
    and t
    val empty : t
    val is_empty : t -> bool
    val mem : elt -> t -> bool
    val add : elt -> t -> t
    val singleton : elt -> t
    val remove : elt -> t -> t
    val union : t -> t -> t
    val inter : t -> t -> t
    val diff : t -> t -> t
    val compare : t -> t -> int
    val equal : t -> t -> bool
    val subset : t -> t -> bool
    val iter : (elt -> unit) -> t -> unit
    val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
    val cardinal : t -> int
    val elements : t -> elt list
    val min_elt : t -> elt
    val max_elt : t -> elt
    val choose : t -> elt
  end
 *)

(* On retourne un top assez violent lors de l'appel de fonctions dont le source
 est inconnu.  Comme ceci inclut toutes les fonctions de bibliotheques
 (malloc(), free(), putchar(), setjmp(), longjmp(), etc.), il faudra raffiner ca un de ces jours. *)

type abs_call_effect =
     (abs_value AbsEnv.t * abs_value option * abs_addr * sig_set) option
	(* L'effet abstrait d'un appel a une fonction:
	soit None si la fonction ne retourne pas,
	soit Some (globs, retval, a, sigs) ou
	globs est la moitie de l'environnement portant sur les valeurs
	des variables globales apres le retour de la fonction,
	retval est la valeur de retour (si pas de type void; sinon,
	retval = None),
	a est une adresse abstraite contenant toutes les adresses sur
	lesquelles l'appel a la fonction a pu faire des effets de bord,
	et sigs est l'ensemble des signaux possiblement lances par
	la fonction. *)

let external_functions = Hashtbl.create 128

let verbose = ref 0

let rec match_params params args ae_loc =
    match params, args with
	[], _ -> ae_loc
      | _, [] -> failwith "match_params"
      | (loc, x, qtyp)::params', av::args' ->
	match_params params' args' (AbsEnv.add x av ae_loc)

exception TestNotTaken
exception FunctionNotFound

let new_finfo () = ref None, ref None, ref z_bot, ref sigs_empty

let rec analyseur (vt : var_table) (ft : fun_table)
    (info : abs_info) (finfo : function_info) (sigs : sig_info) (not_found : functions_not_found)
    (heapsizes : heap_size_info)
    (cctx : CallContext.t) (ctx : loc_ppoint option) (ae_glob : abs_value AbsEnv.t)
    (f : string) (args : abs_value list) (rettyp : q_ctype) =
    (if !verbose>=1
	 then (prerr_string "--> Entering function ";
	       prerr_string f;
	       (match ctx with
		    Some (g, _) -> (prerr_string " [called from ";
				    prerr_string g;
				    prerr_string "]")
		  | None -> ());
	       prerr_newline ()));
    let res =
    try
    let (stat, loc, qtyp, entryp, exitp, cg, n, env, abbrvs) =
	try Hashtbl.find ft f
        with Not_found -> raise FunctionNotFound in
    (match qtyp with
	(_, TFUN (T_PROTOTYPE (params, _), _)) ->
	let initial_pp = (f, entryp), ctx in
	let initial_widen = CallContext.mem initial_pp cctx in
	let cctx' = CallContext.add initial_pp cctx in
	let loc_info = (try Hashtbl.find info ctx
			with Not_found -> let li = Hashtbl.create n in
			     (Hashtbl.add info ctx li; li)) in
	let retae_glob, retval, changed_in_f, sigs_in_f = try
	    let locfi = Hashtbl.find finfo f in
		(try Hashtbl.find locfi ctx
		 with Not_found ->
		      let c = new_finfo () in
			  (Hashtbl.add locfi ctx c;
			   c))
	    with Not_found -> let c = new_finfo () in
		 let locfi = Hashtbl.create 1 in
		     (Hashtbl.add locfi ctx c;
		      Hashtbl.add finfo f locfi;
		      c) in
        let loc_sigs = (try Hashtbl.find sigs ctx
			with Not_found -> let ls = Hashtbl.create n in
			     (Hashtbl.add sigs ctx ls; ls)) in
        let initial_ae = ae_glob, match_params params args (AbsEnv.empty) in
	let worklist = ref ([entryp, initial_ae, initial_widen] :
			      (node * abs_env * bool) list) in
        let rec work () =
	    match !worklist with
		(n, ae, widen)::rest ->
		(worklist := rest;
		let oldae, newae, not_changed =
		    (try let prev_ae = Hashtbl.find loc_info (f, n) in
		     let next_ae = if widen
				       then ae_widen env prev_ae ae
				   else ae_union prev_ae ae in
				       prev_ae, next_ae, ae_equal prev_ae next_ae
			with Not_found -> (AbsEnv.empty, AbsEnv.empty), ae, false) in
		    (if !verbose>=2 && widen
			 then (prerr_string "\tWiden: ";
			       prerr_int n;
			       prerr_string ": ";
			       (if not_changed
				    then prerr_string "-"
				else ae_print_diff ft oldae newae);
				    prerr_newline ()
			    ));
		    (if not_changed
			 then ()
		    else (hash_replace loc_info (f, n) newae; (* !!!should really be Hashtbl.replace *)
			  try let succs = Hashtbl.find cg n in
			      Hashtbl.iter
			      (fun nextn ((_, i) as loci) ->
				   (if !verbose>=2
					then (prerr_string "\t";
					      pr_instr_1 stderr n nextn i)
					    );
				   try
				   let nextae, changed, sgs =
				       (match i with
					    I_SET_VAR (x, y, qtyp) ->
					    ae_assign f ctx newae x (ae_value newae y qtyp) qtyp
					  | I_SET_ARRAY (x, a, qtyp) ->
					    ae_set_array f ctx newae x a qtyp
					  | I_SET_FUN (x, f, qtyp) ->
					    ae_set_fun f ctx newae x f qtyp
					  | I_SET_CST (x, cst, qtyp) ->
					    ae_set_cst f ctx newae x cst qtyp
					  | I_COPY (x, y, qtyp) ->
					    let qtyp_ptr = (false, false), TPOINTER qtyp in
						ae_copy heapsizes f ctx newae (ae_value newae x qtyp_ptr)
						(ae_value newae y qtyp_ptr)
						qtyp
					  | I_PTR (x, y, qtypy, qtypx) ->
					    ae_ptr heapsizes f ctx newae x (ae_value newae y qtypy) qtypy qtypx
					  | I_SET_PTR (x, y, qtypx, qtypy) ->
					    ae_set_ptr heapsizes f ctx newae (ae_value newae x qtypx)
					    (ae_value newae y qtypy) qtypx qtypy
					  | I_ADDR_VAR (x, y, qtyp) ->
					    ae_addr_var f ctx newae x y qtyp
					  | I_ADDR_INDEX (x, y, z, qtyp) ->
					    ae_addr_index f ctx newae x
					    (ae_value newae y ((false, false), TPOINTER qtyp))
					    (ae_value newae z ((false, false), TLONG))
					    qtyp
					  | I_ANTI_ADDR_INDEX (x, y, z, qtyp) ->
					    let qtyp_ptr = (false, false), TPOINTER qtyp in
						ae_anti_addr_index f ctx newae x
						(ae_value newae y qtyp_ptr)
						(ae_value newae z qtyp_ptr)
						qtyp
					  | I_ADDR_SEL (x, y, f, offset, qtypy, qtypx) ->
					    ae_addr_sel f ctx newae x
					    (ae_value newae y qtypy)
					    f offset qtypy qtypx
					  | I_CALL (x, g, args', restyp) ->
					    (match do_func n newae x g args' restyp with
						 Some r -> r
					       | None -> raise TestNotTaken)
					  | I_CALL_IND (x, y, qftyp, args', restyp) ->
					    let yval = ae_value newae y ((false, false),
									   TPOINTER qftyp) in
					    (match yval with
						 ABS_ADDR a ->
						 let funs, danger = z_functions a in
						 let rr = ref [] in
						 (List.iter (fun g ->
								 match do_func n newae x g args' restyp with
								     None -> ()
								   | Some r ->
								     rr := r :: !rr
								 ) funs;
						  if danger
						      then (AbsEnv.empty, AbsEnv.empty),
							    None, all_sigs
						  else match !rr with
							   [] -> raise TestNotTaken
							 | (ae0, a0, sigs0)::rest ->
							   let aer = ref ae0 in
							   let ar = ref a0 in
							   let sigsr = ref sigs0 in
							       (List.iter (fun (ae1, a1, sigs1) ->
									       (aer := ae_union
										(!aer) ae1;
										ar := z_union
										(!ar) a1;
										sigsr := sigs_union
										(!sigsr) sigs1))
								rest;
								!aer, !ar, !sigsr))
						| _ -> failwith "analyseur/I_CALL_IND/not an address"
						 )
					  | I_CONVERT (x, qtypx, qtypy, y) ->
					    ae_convert f ctx newae x qtypx qtypy
					    (ae_value newae y qtypy)
					  | I_CAST (x, qtypx, qtypy, y) ->
					    ae_cast f ctx newae x qtypx qtypy
					    (ae_value newae y qtypy)
					  | I_BNOT (x, y, qtyp) ->
					    ae_bnot f ctx newae x (ae_value newae y qtyp) qtyp
					  | I_MINUS (x, y, qtyp) ->
					    ae_minus f ctx newae x (ae_value newae y qtyp) qtyp
					  | I_MUL (x, y, z, qtyp) ->
					    ae_mul f ctx newae x (ae_value newae y qtyp)
					    (ae_value newae z qtyp) qtyp
					  | I_DIV (x, y, z, qtyp) ->
					    ae_div f ctx newae x (ae_value newae y qtyp)
					    (ae_value newae z qtyp) qtyp
					  | I_MOD (x, y, z, qtyp) ->
					    ae_mod f ctx newae x (ae_value newae y qtyp)
					    (ae_value newae z qtyp) qtyp
					  | I_ADD (x, y, z, qtyp) ->
					    ae_add f ctx newae x (ae_value newae y qtyp)
					    (ae_value newae z qtyp) qtyp
					  | I_LEFT (x, y, z, qtyp) ->
					    ae_left f ctx newae x (ae_value newae y qtyp)
					    (ae_value newae z qtyp) qtyp
					  | I_RIGHT (x, y, z, qtyp) ->
					    ae_right f ctx newae x (ae_value newae y qtyp)
					    (ae_value newae z qtyp) qtyp
					  | I_AND (x, y, z, qtyp) ->
					    ae_and f ctx newae x (ae_value newae y qtyp)
					    (ae_value newae z qtyp) qtyp
					  | I_XOR (x, y, z, qtyp) ->
					    ae_xor f ctx newae x (ae_value newae y qtyp)
					    (ae_value newae z qtyp) qtyp
					  | I_OR (x, y, z, qtyp) ->
					    ae_or f ctx newae x (ae_value newae y qtyp)
					    (ae_value newae z qtyp) qtyp
					  | I_EQ (x, y, z, qtyp) ->
					    ae_eq f ctx newae x (ae_value newae y qtyp)
					    (ae_value newae z qtyp) qtyp
					  | I_LE (x, y, z, qtyp) ->
					    ae_le f ctx newae x (ae_value newae y qtyp)
					    (ae_value newae z qtyp) qtyp
					  | I_LT (x, y, z, qtyp) ->
					    ae_lt f ctx newae x (ae_value newae y qtyp)
					    (ae_value newae z qtyp) qtyp
					  | I_TEST_TRUE x ->
					    if ae_test_true (ae_value newae x ((false, false), TINT))
						then newae, z_bot, sigs_empty
					    else raise TestNotTaken
					  | I_TEST_FALSE x ->
					    if ae_test_false (ae_value newae x ((false, false), TINT))
						then newae, z_bot, sigs_empty
					    else raise TestNotTaken
					  | I_SKIP -> 
					    newae, z_bot, sigs_empty
					    ) in
				       ((if !verbose>=2
					     then (prerr_string "\t";
						   ae_print_diff ft newae nextae;
						   prerr_newline ()));
					worklist := (nextn, nextae, nextn<=n):: !worklist;
					(if nextn=exitp && snd rettyp<>TVOID
					     then let rval = ae_value nextae
						 (".return", false) rettyp in
						 match !retval with
						     None -> retval := Some rval
						   | Some old -> retval := Some (av_union old rval));
					(if nextn=exitp
					     then match !retae_glob with
						      None -> retae_glob := Some (fst nextae)
						    | Some old -> retae_glob :=
						      Some (ae_union_1 old (fst nextae)));
					changed_in_f := z_union (!changed_in_f) changed;
					(if sgs<>sigs_empty
					     then (let oldsgs = try Hashtbl.find loc_sigs loci
						                with Not_found -> sigs_empty in
						   hash_replace loc_sigs loci
						   (sigs_union oldsgs sgs);
						   (*!!!should really be Hashtbl.replace*)
						   sigs_in_f := sigs_union (!sigs_in_f) sgs))
					)
				       with TestNotTaken -> ()
				       ) succs
			  with Not_found -> ()
			  );
			 work ()))
	      | [] -> ()
        and do_func n ae x g args' restyp =
	    (let args_g = List.map (fun (x, qtyp) ->
					ae_value ae x qtyp) args'
		 in
	     match (try let dog = Hashtbl.find external_functions g in
	                    dog vt ft info finfo sigs not_found heapsizes
                                cctx' ((f,n), ctx) (fst ae) args_g restyp
	            with Not_found ->
			 analyseur vt ft info finfo sigs not_found heapsizes
		                  cctx' (Some (f, n)) (fst ae) g args_g restyp) with
		      Some (glob', ret', effect', sigs') ->
		      let loc' = snd (ae_wipe f ctx ae effect') in
			  Some (match ret' with
			        None -> (glob', loc'), effect', sigs'
			      | Some xval -> let newae', a'', sigs'' =
				ae_assign f ctx (glob', loc') x xval restyp in
				newae', z_union effect' a'', sigs_union sigs' sigs'')
                    | None -> None)
        in (work ();
	    match !retae_glob with
		None -> None (* la fin de la fonction n'est pas atteinte *)
	      | Some retae_g -> Some (retae_g, !retval, !changed_in_f, !sigs_in_f))
      | _ -> failwith ("analyseur on " ^ f ^ " with non-functional type")
	)
    with FunctionNotFound ->
	 (hash_replace not_found f ();
	  Some (AbsEnv.empty, (* le top des environnements de variables globales *)
		(match rettyp with
		     (_, TVOID) -> None
		   | _ -> Some (av_top rettyp)), (* le top du type de retour *)
		None, (* fait des effets de bord n'importe ou *)
		all_sigs (* lance tous les signaux possibles *)
		)
	  )
    in
	((if !verbose>=1
	      then (prerr_string "<-- Exiting function ";
		    prerr_string f;
		    prerr_newline ()));
	      res)

(* Note: on peut faire bien plus efficace avec une SSA.
 Pourquoi? *)
(* Note: le resultat doit donner une valeur de retour + une adresse abstraite
 representant toutes les adresses sur lesquelles il a ete effectue une affectation
 (effets de bord).  On pourrait faire plus precis: comment? *)
(* Note: si on modelisait les sauts non locaux (setjmp()/longjmp()), ceci
 ne suffirait pas.  Que proposez-vous? *)
