(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *		use is subject to a specific license, obtainable from LSV.
 *)

open Cparse
open Arith
open Error

type var = string * bool (* son nom, et un booleen [true si globale, false si locale] *)

type instr =
    I_SET_VAR of var * var * q_ctype (* x:=y : type;
	le type est un type entier, flottant ou pointeur ou tableau,
	jamais fonction, struct, union ou void. *)
  | I_SET_ARRAY of var * var * q_ctype (* x:=a : type, ou a est un nom de tableau,
					et type est un TARRAY (...). *)
  | I_SET_FUN of var * string * q_ctype (* x:=f : type, ou f est un nom de fonction,
					 et type est un TFUN (...). *)
  | I_SET_CST of var * constant_val * q_ctype (* x:=constante : type *)
  | I_COPY of var * var * q_ctype (* *x=*y : type (copie de struct ou d'union) *)
  | I_PTR of var * var * q_ctype * q_ctype
  (* I_PTR (x, y, type-de-y, type-de-x): effectue x:=*y;
   le type-de-y est pointeur sur le type-de-x, qui est un type entier, flottant
   ou pointeur lui-meme. *)
  | I_SET_PTR of var * var * q_ctype * q_ctype
  (* I_SET_PTR (x, y, type-de-x, type-de-y): effectue *x:=y;
   le type-de-x est pointeur sur le type-de-y, qui est un type entier, flottant
   ou pointeur lui-meme. *)
  | I_ADDR_VAR of var * var * q_ctype
  (* I_ADDR_VAR (x, y, t): effectue x := &y;
   y est de type t, et x est de type (t * ). *)
  | I_ADDR_INDEX of var * var * var * q_ctype
  (* I_ADDR_INDEX (x, y, z, t): effectue x:=&y[z];
   y est de type pointeur vers t, ainsi que x;
   z est toujours de type long, possiblement const ou volatile. *)
  | I_ANTI_ADDR_INDEX of var * var * var * q_ctype
  (* I_ANTI_ADDR_INDEX (x, y, z, t): effecture x:=y-z;
     y et z sont de meme type (t * ), x est de type long. *)
  | I_ADDR_SEL of var * var * string * int * q_ctype * q_ctype
  (* I_ADDR_SEL (x, y, f, offset, type-de-y, type-de-x): effectue x:=&y->f,
     ou offset est l'offset du champ f par rapport au debut de la structure y. *)
  | I_CALL of var * string * (var * q_ctype) list * q_ctype (* x:=f(x1,...,xn) : type *)
  | I_CALL_IND of var * var * q_ctype * (var * q_ctype) list * q_ctype
  (* I_CALL_IND (x, y, type-de-y, [x1;...;xn], type-retour)
   effectue x:= ( *y)(x1,...,xn) : type-retour *)
  | I_CONVERT of var * q_ctype * q_ctype * var
  (* I_CONVERT (x, T1, T2, y): x:=(T1)y avec y:T2;
   il s'agit d'une conversion (int -> double, etc.), qui implique un changement de
   representation, contrairement au cast, qui est juste un changement de type.
   On a uniquement les cas suivants:
   float <-> double
   long <-> double
   long <-> unsigned long
   int <-> unsigned int
   short <-> unsigned short
   char <-> unsigned char
   long <-> int
   unsigned long <-> unsigned int
   long <-> short
   unsigned long <-> unsigned short
   long <-> char
   unsigned long <-> unsigned char
   int <-> short
   unsigned int <-> unsigned short
   int <-> char
   unsigned int <-> unsigned char
   short <-> char
   unsigned short <-> unsigned char
   unsigned long <-> T * (pour n'importe quel type T)
   T [...] -> T * (on peut convertir un type tableau en un type pointeur correspondant,
                   mais pas le contraire)
   *)
  | I_CAST of var * q_ctype * q_ctype * var (* x:=(T1)y avec y:T2 *)
  | I_BNOT of var * var * q_ctype (* x:=~y : type *)
  (* Le type est char, short, int, ou long, possiblement unsigned,
   possiblement const ou volatile.  On notera que
   ~y = -y-1 modulo les bornes de definition du type entier considere. *)
  | I_MINUS of var * var * q_ctype (* x:=-y : type (type est un type numerique,
				    int, char, short, long
				    [possiblement unsigned], float, double;
				    possiblement const ou volatile) *)
  | I_MUL of var * var * var * q_ctype (* x:=y*z *)
    (* multiplication, le type est char, short, int, long, float
     ou double, possiblement unsigned, const ou volatile. *)
  | I_DIV of var * var * var * q_ctype (* x:=y/z *)
    (* division, le type est char, short, int, long, float
     ou double, possiblement unsigned, const ou volatile. *)
  | I_MOD of var * var * var * q_ctype (* x:=y%z *)
    (* reste, le type est char, short, int, ou long,
     possiblement unsigned, const ou volatile. *)
  | I_ADD of var * var * var * q_ctype (* x:=y+z *)
    (* addition, le type est char, short, int, long, float
     ou double, possiblement unsigned, const ou volatile. *)
  | I_LEFT of var * var * var * q_ctype (* x:=y<<z *)
    (* le type est celui de x et de y; c'est char, short, int, long, float
     ou double, possiblement unsigned, const ou volatile.
     Le type de z est toujours int. *)
  | I_RIGHT of var * var * var * q_ctype (* x:=y>>z *)
    (* le type est celui de x et de y; c'est char, short, int, long, float
     ou double, possiblement unsigned, const ou volatile.
     Le type de z est toujours int. *)
  | I_AND of var * var * var * q_ctype (* x:=y&z *)
    (* le type est int, unsigned int, long ou unsigned long. *)
  | I_XOR of var * var * var * q_ctype (* x:=y^z *)
    (* le type est char, short, int, long, possiblement unsigned, const ou volatile. *)
  | I_OR of var * var * var * q_ctype (* x:=y|z *)
    (* le type est char, short, int, long, possiblement unsigned, const ou volatile. *)
  | I_EQ of var * var * var * q_ctype (* x:=(y==z) avec y,z:type *)
    (* le type est char, short, int, long, possiblement unsigned,
     ou float ou double, possiblement const ou volatile,
     ou bien un type non numerique (pointeur, tableau).
     Le type de x est toujours int. *)
  | I_LE of var * var * var * q_ctype (* x:=(y<=z) avec y,z:type *)
    (* le type en question est char, short, int, long, float
     ou double, possiblement unsigned, const ou volatile.
     Le type de x est toujours int. *)
  | I_LT of var * var * var * q_ctype (* x:=(y<z) avec y,z:type *)
    (* le type en question est char, short, int, long, float
     ou double, possiblement unsigned, const ou volatile.
     Le type de x est toujours int. *)
  | I_TEST_TRUE of var (* transition prise si x est vraie (entier non nul) *)
  | I_TEST_FALSE of var (* transition prise si x est fausse (entier nul) *)
  | I_SKIP (* ne fait rien. *)

type loc_instr = locator * instr

type node = int

type control_graph = (node, (node, loc_instr) Hashtbl.t) Hashtbl.t
    (* envoie chaque sommet [un int] vers une map
     qui envoie chaque successeur [un int encore]
     vers une instruction.
     Il s'agit du graphe de controle d'une seule
     fonction. *)

(* Un point de programme local: le nom de la fonction
 dans laquelle on est, plus le sommet [un int] dans le
 graphe de controle associe. *)
type loc_ppoint = string * node

(* Un (vrai) point de programme =
 un point de programme local (la ou on est),
 plus un autre point de programme local (d'ou on a ete appele)
 ou None si on est dans main(). *)
type ppoint = loc_ppoint * loc_ppoint option

let gennode_count = ref 0
let gennode () = (incr gennode_count; !gennode_count)

let genvar loc_e =
    (* match loc_e with
        (_, VAR (x, _)) -> x
      | _ -> *) gensym ".x"

let cg_addnode cg src tgt instr =

(* DEBUG *) let _ = Debug.pRINT_DEBUG2 "cg_addnode : main" in

    try let succs = Hashtbl.find cg src in
        Hashtbl.add succs tgt instr
    with Not_found -> Hashtbl.add cg src (let succs = Hashtbl.create 1 in
					      (Hashtbl.add succs tgt instr;
					       succs))
(*
    match map_read cg src with
        Some succs ->
	(match map_read succs tgt with
	     Some _ -> failwith ("cg_addnode: deja une instruction de " ^
				 string_of_int src ^ " vers " ^ string_of_int tgt)
	   | None -> map_add cg src (map_add succs tgt instr))
      | None -> map_add cg src (map_singleton tgt instr)
*)

let rec side_effecting_exprs ((_, _, e) as loce) acc =
    match e with
        VAR (_, _) -> acc
      | ARRAY (_, _) -> acc
      | FUN _ -> acc
      | CST _ -> acc
      | PTR a -> side_effecting_exprs a acc
      | SET_VAR (x, e', _) -> side_effecting_exprs e' (loce::acc)
      | SET_PTR (a, e') ->
	side_effecting_exprs a (side_effecting_exprs e' (loce::acc))
      | COPY (e, e') ->
	side_effecting_exprs e (side_effecting_exprs e' (loce::acc))
      | ADDR_VAR (_, _) -> acc
      | ADDR_INDEX (a, i) -> side_effecting_exprs a (side_effecting_exprs i acc)
      | ANTI_ADDR_INDEX (a, b) -> side_effecting_exprs a (side_effecting_exprs b acc)
      | ADDR_SEL (a, f, _) -> side_effecting_exprs a acc
      | CALL (_, fl) -> side_effecting_exprs_l fl (loce::acc)
      | CALL_IND (f, _, fl) -> side_effecting_exprs f (side_effecting_exprs_l fl (loce::acc))
      | CAST e' -> side_effecting_exprs e' acc
      | CONVERT e' -> side_effecting_exprs e' acc
      | BNOT e' -> side_effecting_exprs e' acc
      | MINUS e' -> side_effecting_exprs e' acc
      | OP (_, e', e'') -> side_effecting_exprs e' (side_effecting_exprs e'' acc)
      | CMP (_, e', e'') -> side_effecting_exprs e' (side_effecting_exprs e'' acc)
      | EIF (e1, e2, e3) ->
	side_effecting_exprs e1 (side_effecting_exprs e2 (side_effecting_exprs e3 acc))
      | ESEQ l -> side_effecting_exprs_l l acc
and side_effecting_exprs_l l acc =
    match l with
        [] -> acc
      | e::l' -> side_effecting_exprs_l l' (side_effecting_exprs e acc)

let check_side_effecting_exprs ((loc, _, _) as loce) =
    let l = side_effecting_exprs loce [] in
    match l with
        _ :: _ :: _ -> let buf = Buffer.create 128 in
	               (Buffer.add_string buf "expression involves doing side effects in some indeterminate order (one will be chosen arbitrarily): ";
			bufout_loc_expr_list buf l;
			warning (Some loc) (Buffer.contents buf))
     | _ -> ()

let rec cg_of_expr_0 cg abbrvs p1 p2 x ((loc, qtyp, e) as loce) =
    (* fabrique un sous-graphe de controle entre p1 et p2 representant
     l'execution de x = e (en fait, UNE execution, il peut y en avoir
     plusieurs---mais ce sera detecte par check_side_effecting_exprs). *)
    match e with
	VAR (y, glob) ->
          begin
            (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "cg_of_expr_0 : VAR" in
	    cg_addnode cg p1 p2 (loc, if fst x=y then I_SKIP else I_SET_VAR (x, (y, glob), qtyp))
          end
      | ARRAY (a, glob) ->
	cg_addnode cg p1 p2 (loc, I_SET_ARRAY (x, (a, glob), qtyp))
      | FUN f ->
	cg_addnode cg p1 p2 (loc, I_SET_FUN (x, f, qtyp))
      | CST c -> cg_addnode cg p1 p2 (loc, I_SET_CST (x, c, qtyp))
      | PTR a -> let ax, ap = cg_of_expr_1 cg abbrvs p1 a in
	cg_addnode cg ap p2 (loc, I_PTR (x, ax,
					 qtyp_of_expr a,
					 qtyp))
      | SET_VAR (y, e', glob) ->
          begin
            (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "cg_of_expr_0 : SET_VAR " in

	    if fst x=y then cg_of_expr_0 cg abbrvs p1 p2 x e' else
	    let ap = gennode () in

            (* DEBUG *) let _ = Debug.pRINT_DEBUG2 ("gennode for node number : "^(string_of_int ap)) in

	    (cg_of_expr_0 cg abbrvs p1 ap (y, glob) e';
	     cg_addnode cg ap p2 (loc, I_SET_VAR (x, (y, glob), qtyp)))
          end
      | SET_PTR (a, e') -> let ax, ap = cg_of_expr_1 cg abbrvs p1 a in
			     let ep = gennode () in
				 (cg_of_expr_0 cg abbrvs ap ep x e';
				  cg_addnode cg ep p2 (loc, I_SET_PTR (ax, x,
								       qtyp_of_expr a,
								       qtyp)))
      | COPY (e, e') -> let ex, ep = cg_of_expr_1 cg abbrvs p1 e in
	                let e'x, e'p = cg_of_expr_1 cg abbrvs ep e' in
			    cg_addnode cg e'p p2 (loc, I_COPY (ex, e'x, qtyp))
      | ADDR_VAR (y, glob) ->
          begin
            (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "cg_of_expr_0 : ADDR_VAR" in
	    match qtyp with
	      (_, TPOINTER qtypy) ->
	        cg_addnode cg p1 p2 (loc, I_ADDR_VAR (x, (y, glob), qtypy))
	    | _ -> failwith "cg_of_expr_0, case ADDR_VAR, is not a TPOINTER"
          end
      | ADDR_INDEX (a, i) -> (let ax, ap = cg_of_expr_1 cg abbrvs p1 a in
	                      let ix, ip = cg_of_expr_1 cg abbrvs ap i in
				 match qtyp_of_expr a with
				     (_, TPOINTER qtyp') ->
				     cg_addnode cg ip p2 (loc,
							  I_ADDR_INDEX (x, ax, ix,
									qtyp'))
				   | _ -> failwith "in ADDR_INDEX, expected TPOINTER type")
      | ANTI_ADDR_INDEX (a, b) -> (let ax, ap = cg_of_expr_1 cg abbrvs p1 a in
	                           let bx, bp = cg_of_expr_1 cg abbrvs ap b in
				      match qtyp_of_expr a with
					  (_, TPOINTER qtyp') ->
					  cg_addnode cg bp p2 (loc,
							       I_ANTI_ADDR_INDEX (x, ax, bx,
										  qtyp'))
					| _ -> failwith "in ANTI_ADDR_INDEX, expected TPOINTER type")
      | ADDR_SEL (a, f, offset) -> let ax, ap = cg_of_expr_1 cg abbrvs p1 a in
	                           cg_addnode cg ap p2 (loc, I_ADDR_SEL (x, ax, f, offset,
									 qtyp_of_expr a,
									 qtyp))
      | CALL (fname, args) -> let xl, xp = cg_of_expr_l cg abbrvs p1 [] args in
	                          cg_addnode cg xp p2 (loc, I_CALL (x, fname,
								    List.rev xl, qtyp))
      | CALL_IND (locf, qftyp, args) ->
	let fx, fp = cg_of_expr_1 cg abbrvs p1 locf in
	let xl, xp = cg_of_expr_l cg abbrvs fp [] args in
	    cg_addnode cg xp p2 (loc, I_CALL_IND (x, fx, qftyp, List.rev xl, qtyp))
      | CONVERT a -> let ax, ap = cg_of_expr_1 cg abbrvs p1 a in
	             cg_addnode cg ap p2 (loc, I_CONVERT (x, qtyp, qtyp_of_expr a, ax))
      | CAST a -> let ax, ap = cg_of_expr_1 cg abbrvs p1 a in
	              cg_addnode cg ap p2 (loc, I_CAST (x, qtyp, qtyp_of_expr a, ax))
      | BNOT a -> let ax, ap = cg_of_expr_1 cg abbrvs p1 a in
	              cg_addnode cg ap p2 (loc, I_BNOT (x, ax, qtyp))
      | MINUS a -> let ax, ap = cg_of_expr_1 cg abbrvs p1 a in
	               cg_addnode cg ap p2 (loc, I_MINUS (x, ax, qtyp))
      | OP (setop, e1, e2) -> let ex1, ep1 = cg_of_expr_1 cg abbrvs p1 e1 in
	                      let ex2, ep2 = cg_of_expr_1 cg abbrvs ep1 e2 in
			      let instr = match setop with
					      S_MUL -> I_MUL (x, ex1, ex2, qtyp)
					    | S_DIV -> I_DIV (x, ex1, ex2, qtyp)
					    | S_MOD -> I_MOD (x, ex1, ex2, qtyp)
					    | S_ADD -> I_ADD (x, ex1, ex2, qtyp)
					    | S_LEFT -> I_LEFT (x, ex1, ex2, qtyp)
					    | S_RIGHT -> I_RIGHT (x, ex1, ex2, qtyp)
					    | S_AND -> I_AND (x, ex1, ex2, qtyp)
					    | S_XOR -> I_XOR (x, ex1, ex2, qtyp)
					    | S_OR -> I_OR (x, ex1, ex2, qtyp)
					    | S_EQ -> failwith "cg_of_expr: setop='='"
					    | S_SUB -> failwith "cg_of_expr: setop='-'" in
				  cg_addnode cg ep2 p2 (loc, instr)
     | CMP (cmpop, e1, e2) -> let ex1, ep1 = cg_of_expr_1 cg abbrvs p1 e1 in
	                      let ex2, ep2 = cg_of_expr_1 cg abbrvs ep1 e2 in
			      let instr = match cmpop with
					      C_LT -> I_LT (x, ex1, ex2, qtyp_of_expr e1)
					    | C_LE -> I_LE (x, ex1, ex2, qtyp_of_expr e1)
					    | C_EQ -> I_EQ (x, ex1, ex2, qtyp_of_expr e1) in
				  cg_addnode cg ep2 p2 (loc, instr)
     | EIF (a, b, c) ->
       (match a with
	    (loca, _, EIF (aif, athen, aelse)) ->
	    cg_of_expr_0 cg abbrvs p1 p2 x (loc, qtyp, EIF (aif,
							    (loc, qtyp, EIF (athen, b, c)),
							    (loc, qtyp, EIF (aelse, b, c))))
          | (_, _, CST (INT n)) ->
	    if float_is_zero n
		then cg_of_expr_0 cg abbrvs p1 p2 x c
	    else cg_of_expr_0 cg abbrvs p1 p2 x b
	  | _ -> let ax, ap = cg_of_expr_1 cg abbrvs p1 a in
	         let thenp = gennode () in
		 let elsep = gennode () in
		     (cg_addnode cg ap thenp (loc, I_TEST_TRUE ax);
		      cg_of_expr_0 cg abbrvs thenp p2 x b;
		      cg_addnode cg ap elsep (loc, I_TEST_FALSE ax);
		      cg_of_expr_0 cg abbrvs elsep p2 x c))
     | ESEQ [] -> cg_addnode cg p1 p2 (loc, I_SKIP)
     | ESEQ l -> cg_of_expr_s cg abbrvs p1 p2 x l
and cg_of_expr_1 cg abbrvs p1 loce =
    let p2 = gennode () in
	cg_of_expr_2 cg abbrvs p1 p2 loce, p2

and cg_of_expr_2 cg abbrvs p1 p2 loce =
  
    (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "Appel de cg_of_expr_2 " in
    (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "Loce = " in
    (* DEBUG *) let _ = Debug.pRINT_DEBUG2 ("Loce = "^(string_of_expr loce)) in

    match loce with
        (_, _, VAR (x, glob)) -> 
          begin
             (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "Appel de cg_of_expr_2 : cas VAR" in

            cg_of_expr_0 cg abbrvs p1 p2 (x, glob) loce;
	    (x, glob)
          end
      | _ -> 
(* DEBUG *) let _ = Debug.pRINT_DEBUG2 "Appel de cg_of_expr_2 : cas <> VAR" in
          let x = genvar loce in
          
          (* DEBUG *) let _  =  Debug.pRINT_DEBUG2 ("Gnration de variable :"^x) in 
		(Hashtbl.add abbrvs x loce;
		 cg_of_expr_0 cg abbrvs p1 p2 (x, false) loce;
		 x, false)
          

and cg_of_expr_l cg abbrvs p1 acc l =
    match l with
        [] -> acc, p1
      | (loce::rest) -> let ex, ep = cg_of_expr_1 cg abbrvs p1 loce in
	                cg_of_expr_l cg abbrvs ep ((ex, qtyp_of_expr loce)::acc) rest

and cg_of_expr_s cg abbrvs p1 p2 x l =
    match l with
        [e] -> cg_of_expr_0 cg abbrvs p1 p2 x e
      | (e::rest) -> let ex, ep = cg_of_expr_1 cg abbrvs p1 e in
	             cg_of_expr_s cg abbrvs ep p2 x rest
      | [] -> failwith "cg_of_expr_s: l=[]"

let cg_of_expr cg abbrvs p1 p2 x loce =
    ((* check_side_effecting_exprs loce; *)
	 (*!!!note: les fonctions ci-dessous utilisent tout sauf cg_of_expr;
	  si on voulait etre serieux, on ferait des versions appelant
	  check_side_effecting_exprs pour cg_of_expr_1 et cg_of_expr_2;
	  d'un autre cote, ne serait-ce que *to++ = *from++ va generer des
	  warnings... a moins qu'on puisse assurer qu'il n'y a pas d'alias entre
	  &to, &from, et to. *)
     cg_of_expr_0 cg abbrvs p1 p2 x loce)

type switch_info = locator * var * constant_val list ref * bool ref
    * (loc_instr * node) ref
    (* le locator de l'expression sur laquelle on fait le switch,
     la variable x representant l'expression sur laquelle on fait le switch,
     la liste des constantes c telles qu'on a deja vu 'case c',
     un booleen disant si on a deja vu le cas 'default',
     et le predecesseur du dernier IF_TRUE, ou le sommet du graphe de controle
     juste apres le calcul de x si pas de dernier IF_TRUE. *)

let cg_of_code funname code =
  
    (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "Appel de cg_of_code" in

    let abbrvs = (Hashtbl.create 256 : (string, loc_expr) Hashtbl.t) in
    let cg = (Hashtbl.create 256 : control_graph) in
    let labels = (Hashtbl.create 10 : (string, node) Hashtbl.t) in
    let gotos = ref ([] : (locator * node * string) list) in
    let entryp = gennode () in
    let returnp = gennode () in
    let retvar = ".return", false in
    let rec make (breaks : node list) (conts : node list) (sws : switch_info list) (loc, c) p1 p2 =
	match c with
            CLABEL (lab, c') ->
	    (
              (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "Appel de cg_of_code : CLABEL" in
             (try (let _ = Hashtbl.find labels lab in
		   error (Some loc) ("duplicate label " ^ lab))
	      with Not_found -> Hashtbl.add labels lab p1);
		 make breaks conts sws c' p1 p2)
	  | CCASE (loc', cst, c') ->
	    (* Note: les CSWITCH ne sont pas necessairement juste au-dessus des CCASE
	     correspondants.  Par exemple, le code suivant [Duff's device, si je me
							     souviens bien],
	     compile parfaitement:
	     main (int argc, char *from, char *to)
	     {
	       int i;
	       int n = argc/8;

	       i = 0;
	       switch (argc % 8) {
                 for (; i<n; i++)
		   {
		     case 0: *to++ = *from++;
		     case 1: *to++ = *from++;
		     case 2: *to++ = *from++;
		     case 3: *to++ = *from++;
		     case 4: *to++ = *from++;
		     case 5: *to++ = *from++;
		     case 6: *to++ = *from++;
		     case 7: *to++ = *from++;
		  }
		}
	     }
	     *)
	    ( 
              (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "Appel de cg_of_code : CCASE" in
              match sws with
		 (locx, x, cases, def, prevr) :: _ ->
		 if !def
		     then (error (Some loc') "case after 'default'";
			   make breaks conts sws c' p1 p2)
		 else ((if List.mem cst (!cases)
			    then error (Some loc') "duplicate case"
		        else ());
			    let (testprev, prevp) = !prevr in
			    let cmp = (loc', ((false, false), TINT),
				       CMP (C_EQ, (locx, ((false, false), TINT), VAR (fst x, snd x)),
					    (loc', ((false, false), TINT), CST cst))) in
			    let pp = gennode () in
			        (cg_addnode cg prevp pp testprev;
 				 let ex, ep = cg_of_expr_1 cg abbrvs pp cmp in
				     (cg_addnode cg ep p1 (loc', I_TEST_TRUE ex);
				      prevr := ((loc', I_TEST_FALSE ex), ep);
				      cases := cst :: !cases;
				      make breaks conts sws c' p1 p2)))
               | [] -> (error (Some loc') "case not within a switch";
		        make breaks conts sws c' p1 p2))
          | CDEFAULT c' ->
	    (
             (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "Appel de cg_of_code : CDEFAULT" in
             match sws with
                 (locx, x, cases, def, prevr) :: _ ->
		 if !def
		     then (error (Some loc) "duplicate 'default'";
			   make breaks conts sws c' p1 p2)
		 else let (testprev, prevp) = !prevr in
		     (cg_addnode cg prevp p1 testprev;
		      def := true;
		      make breaks conts sws c' p1 p2)
               | [] -> (error (Some loc) "'default' not within a switch";
		        make breaks conts sws c' p1 p2))
          | CBLOCK (_, codel) ->
	    (
             (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "Appel de cg_of_code : CBLOCK" in
             match codel with
	         [] -> cg_addnode cg p1 p2 (loc, I_SKIP)
               | _ -> make_l breaks conts sws codel p1 p2)
          | CEXPR loce -> 
              begin
                (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "Appel de cg_of_code : CEXPR" in
                ignore (cg_of_expr_2 cg abbrvs p1 p2 loce)
              end
          | CIF (a, c1, c2) ->
	    (
             (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "Appel de cg_of_code : CIF" in
             match a with
		 (loca, _, EIF (aif, athen, aelse)) ->
		 make breaks conts sws (loc, CIF (aif,
					    (loc, CIF (athen, c1, c2)),
					    (loc, CIF (aelse, c1, c2)))) p1 p2
	       | (_, _, CST (INT n)) ->
		 if float_is_zero n
		     then make breaks conts sws c2 p1 p2
		 else make breaks conts sws c1 p1 p2
	       | _ -> let ax, ap = cg_of_expr_1 cg abbrvs p1 a in
		      let thenp = gennode () in
		      let elsep = gennode () in
			  (cg_addnode cg ap thenp (loc, I_TEST_TRUE ax);
			   make breaks conts sws c1 thenp p2;
			   cg_addnode cg ap elsep (loc, I_TEST_FALSE ax);
			   make breaks conts sws c2 elsep p2))
          | CSWITCH ((loc, _, _) as loce, c) ->
              begin
                (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "Appel de cg_of_code : CSWITCH" in
	        let ex, ep = cg_of_expr_1 cg abbrvs p1 loce in
	        let def = ref false in
                let last = ref ((loc, I_SKIP), ep) in
	        let sws' = (loc, ex, ref [], def, last)::sws in
	        let p0 = gennode () in
		(make (p2::breaks) conts sws' c p0 p2;
		 if !def
		   then ()
		   else let (testprev, prevp) = !last in
		            cg_addnode cg prevp p2 testprev)
              end
          | CLOOP (c1, c2) ->
	    (* boucler sur c1; c2;
	     on demarre la boucle juste apres c1;
	     le label continue est juste apres c2. *)
  (* DEBUG *) let _ =  Debug.pRINT_DEBUG2 "Appel de cg_of_code : CLOOP" in
	    let p0 = gennode () in
	    let pstart = gennode () in
	    let breaks' = p2::breaks in
	    let conts' = p0::conts in
		(cg_addnode cg p1 pstart (loc, I_SKIP);
		 make breaks' conts' sws c1 p0 pstart;
		 make breaks' conts' sws c2 pstart p0)
	  | CGOTO lab -> 
              begin
                (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "Appel de cg_of_code : CGOTO" in
                gotos := (loc, p1, lab) :: !gotos
              end
          | CCONTINUE ->
	    (
             (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "Appel de cg_of_code : CCONTINUE" in
             match conts with
		 [] -> (error (Some loc) "continue statement not within a loop")
	       | p :: _ -> cg_addnode cg p1 p (loc, I_SKIP))
	  | CBREAK ->
	    (
             (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "Appel de cg_of_code : CBREAK " in
             match breaks with
		 [] -> (error (Some loc) "break statement not within loop or switch")
	       | p :: _ -> cg_addnode cg p1 p (loc, I_SKIP))
	  | CRETURN None -> 
              begin
                (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "Appel de cg_of_code : CRETURN none" in
                cg_addnode cg p1 returnp (loc, I_SKIP)
              end
	  | CRETURN (Some loce) ->
              begin
                (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "Appel de cg_of_code : CRETURN Some ()" in
	        cg_of_expr cg abbrvs p1 returnp retvar loce
              end
        and make_l breaks conts sws codel p1 p2 =
	    (match codel with
		 [] -> failwith "make_l on empty list"
	       | [c] -> make breaks conts sws c p1 p2
	       | c::l -> let p = gennode () in
		         (make breaks conts sws c p1 p;
			  make_l breaks conts sws l p p2))
    in (make [] [] [] code entryp returnp;
	List.iter (fun (loc, p, lab) ->
		       (try let p' = Hashtbl.find labels lab in
			    cg_addnode cg p p' (loc, I_SKIP)
			with Not_found -> error (Some loc) ("label " ^ lab ^ " undefined")
		       )) (!gotos);
	entryp, returnp, cg, abbrvs)

exception Pursue of node

let cg_rev_top_order cg entryp =
    let hash = Hashtbl.create 128 in
    let lr = ref [] in
    let rec cg_rto i =
	try ignore (Hashtbl.find hash i)
        with Not_found -> (Hashtbl.add hash i ();
			   (try let succs = Hashtbl.find cg i in
			        Hashtbl.iter (fun j _ -> cg_rto j) succs
			    with Not_found -> ());
			   lr := i :: !lr) in
    let rec pursue () =
	try Hashtbl.iter (fun i _ -> try ignore (Hashtbl.find hash i)
			      with Not_found -> raise (Pursue i)) cg
	with Pursue i -> (cg_rto i; pursue ())
    in (cg_rto entryp; pursue (); !lr)

(*
let hash_copy h = (* pas de Hashtbl.copy en Ocaml 2.04! *)
    let h' = Hashtbl.create 128 in
	(Hashtbl.iter (fun x y -> Hashtbl.add h' x y) h;
	 h')

exception Pursue of node

let cg_rev_top_order cg entryp =
    let cg' = hash_copy cg in
    let lr = ref [] in
    let rec cg_rto i =
	try let succs = Hashtbl.find cg' i in
	        (Hashtbl.remove cg' i;
		 Hashtbl.iter (fun j _ -> cg_rto j) succs;
		 lr := i :: !lr)
        with Not_found -> lr := i::!lr in
    let rec pursue () =
	try Hashtbl.iter (fun i _ -> raise (Pursue i)) cg'
	with Pursue i -> (cg_rto i; pursue ())
    in (cg_rto entryp; pursue (); !lr)
*)

let cg_type_env cg rto params =
    let env = (Hashtbl.create 128 : (var, q_ctype) Hashtbl.t) in
    let var x qtyp = try let qtyp' = Hashtbl.find env x in
	                     if snd qtyp<>snd qtyp'
				 then 
                               begin
                                 
                                 let s1 = string_of_qtyp qtyp in
                                 let s2 = string_of_qtyp qtyp' in

                                 (* DEBUG *) let _ = Debug.pRINT_WARNING ("cg_type_env: " ^ (fst x) ^ " declared twice with different types : "^s1 ^" and "^s2) in
                                 (* DEBUG *) (*                                
                                  failwith ("cg_type_env: " ^ (fst x) ^ " declared twice with different types : "^s1 ^" and "^s2) ;
                                 *) (* DEBUG *)
                                 ()
                               end
	             with Not_found -> Hashtbl.add env x qtyp in
    (List.iter (fun (loc, x, qtyp) -> var (x, false) qtyp) params;
     List.iter (fun i ->
		    try let succs = Hashtbl.find cg i in
		        Hashtbl.iter (fun _ (loc, instr) ->
					  match instr with
					      I_SET_VAR (x, _, qtyp) -> var x qtyp
					    | I_SET_ARRAY (x, _, qtyp) -> var x qtyp
					    | I_SET_FUN (x, _, qtyp) -> var x qtyp
					    | I_SET_CST (x, _, qtyp) -> var x qtyp
					    | I_PTR (x, _, _, qtyp) -> var x qtyp
					    | I_ADDR_VAR (x, _, qtyp) -> var x ((false, false), TPOINTER qtyp)
					    | I_ADDR_INDEX (x, _, _, qtyp) -> var x ((false, false), TPOINTER qtyp)
					    | I_ANTI_ADDR_INDEX (x, _, _, _) -> var x ((false, false), TLONG)
					    | I_ADDR_SEL (x, _, _, _, _, qtyp) -> var x qtyp
					    | I_CALL (x, _, _, qtyp) -> var x qtyp
					    | I_CALL_IND (x, _, _, _, qtyp) -> var x qtyp
					    | I_CONVERT (x, qtyp, _, _) -> var x qtyp
					    | I_CAST (x, qtyp, _, _) -> var x qtyp
					    | I_BNOT (x, _, qtyp) -> var x qtyp
					    | I_MINUS (x, _, qtyp) -> var x qtyp
					    | I_MUL (x, _, _, qtyp) -> var x qtyp
					    | I_DIV (x, _, _, qtyp) -> var x qtyp
					    | I_MOD (x, _, _, qtyp) -> var x qtyp
					    | I_ADD (x, _, _, qtyp) -> var x qtyp
					    | I_LEFT (x, _, _, qtyp) -> var x qtyp
					    | I_RIGHT (x, _, _, qtyp) -> var x qtyp
					    | I_AND (x, _, _, qtyp) -> var x qtyp
					    | I_XOR (x, _, _, qtyp) -> var x qtyp
					    | I_OR (x, _, _, qtyp) -> var x qtyp
					    | I_EQ (x, _, _, _) -> var x ((false, false), TINT)
					    | I_LE (x, _, _, _) -> var x ((false, false), TINT)
					    | I_LT (x, _, _, _) -> var x ((false, false), TINT)
					    | _ -> ()
					  ) succs
		    with Not_found -> ()) rto;
     env)

type node_ranking = (node, int) Hashtbl.t

    (* fabrique une numerotation des noeuds d'un graphe de controle
     de sorte que les arcs arriere m -> n soient tels que m<=n.
     On exploite le fait qu'on a deja un tri topologique inverse
     (en ordre inverse) calcule dans les Control.fun_table.
     Les numeros m, n des noeuds sont negatifs, mais que diable... *)
let get_ranking node_list =
    let rank = Hashtbl.create 128 in
    let nr = ref 0 in
	(List.iter (fun node -> (Hashtbl.add rank node (!nr);
				 decr nr)) node_list;
	 rank, !nr)

let rank_number ranking minn n =
    try 1-Hashtbl.find ranking n (* Hashtbl.find ranking n - minn *)
    with Not_found -> failwith "rank_number"

let cg_renumber cg ranking minn =
    let cg' = Hashtbl.create 128 in
	(Hashtbl.iter (fun src succs ->
			   let src' = rank_number ranking minn src in
			       let succs' = Hashtbl.create 1 in
				   (Hashtbl.add cg' src' succs';
				    Hashtbl.iter (fun tgt li ->
						      let tgt' = rank_number ranking minn tgt in
							  Hashtbl.add succs' tgt' li) succs)) cg;
	 cg')

let output_int file i =
    output_string file (string_of_int i)

let output_float file x =
    output_string file (string_of_float x)

let pr_cst file cst =
    match cst with
        INT    x -> output_float file x
      | FLOAT  x -> output_float file x
      | STRING s -> output_string file (String.escaped s)

let pr_args file args =
    match args with
	[] -> output_string file "()"
      | ((x, _),qtyp)::rest ->
	(output_string file "(";
	 output_string file x;
	 output_string file " : ";
	 pr_qtype file qtyp;
	 List.iter (fun ((x, _), qtyp) ->
			(output_string file ", ";
			 output_string file x;
			 output_string file " : ";
			 pr_qtype file qtyp)) rest;
	 output_string file ")")

let pr_instr file instr =
    match instr with
	I_SET_VAR ((x, _), (y, _), qtyp) -> (output_string file "I_SET_VAR ";
                                             output_string file x;
					     output_string file ":=";
					     output_string file y;
					     output_string file "\t: ";
					     pr_qtype file qtyp)
      | I_SET_ARRAY ((x, _), (a, _), qtyp) -> (output_string file "I_SET_ARRAY ";
                                               output_string file x;
					       output_string file ":=";
					       output_string file a;
					       output_string file " [array]\t: ";
					       pr_qtype file qtyp)
      | I_SET_FUN ((x, _), f, qtyp) -> (output_string file "I_SET_FUN ";
                                        output_string file x;
					output_string file ":=";
					output_string file f;
					output_string file " [function]\t: ";
					pr_qtype file qtyp)
      | I_SET_CST ((x, _), cst, qtyp) -> (output_string file "I_SET_CST ";
                                          output_string file x;
					  output_string file ":=";
					  pr_cst file cst;
					  output_string file "\t: ";
					  pr_qtype file qtyp)
      | I_COPY ((x, _), (y, _), qtyp) -> (output_string file "I_COPY ";
                                          output_string file "*";
					  output_string file x;
					  output_string file ":=*";
					  output_string file y;
					  output_string file "\t: ";
					  pr_qtype file qtyp)
      | I_PTR ((x, _), (y, _), qtypy, qtypx) ->
	(output_string file "I_PTR ";
         output_string file x;
	 output_string file ":=*";
	 output_string file y;
	 output_string file "\t; ";
	 output_string file x;
	 output_string file " : ";
	 pr_qtype file qtypx;
	 output_string file ", ";
	 output_string file y;
	 output_string file " : ";
	 pr_qtype file qtypy)
      | I_SET_PTR ((x, _), (y, _), qtypx, qtypy) ->
	(output_string file "I_SET_PTR ";
         output_string file "*";
	 output_string file x;
	 output_string file ":=";
	 output_string file y;
	 output_string file "\t; ";
	 output_string file x;
	 output_string file " : ";
	 pr_qtype file qtypx;
	 output_string file ", ";
	 output_string file y;
	 output_string file " : ";
	 pr_qtype file qtypy)
      | I_ADDR_VAR ((x, _), (y, _), qtypy) ->
	(output_string file "I_ADDR_VAR ";
         output_string file x;
	 output_string file ":=&";
	 output_string file y;
	 output_string file "; ";
	 output_string file x;
	 output_string file " : ";
	 pr_qtype file qtypy;
	 output_string file " *")
      | I_ADDR_INDEX ((x, _), (y, _), (z, _), qtyp) ->
	(output_string file "I_ADDR_INDEX ";
         output_string file x;
	 output_string file ":=&";
	 output_string file y;
	 output_string file "[";
	 output_string file z;
	 output_string file "]\t; ";
	 output_string file x;
	 output_string file ", ";
	 output_string file y;
	 output_string file " : ";
	 pr_qtype file qtyp;
	 output_string file " *")
      | I_ANTI_ADDR_INDEX ((x, _), (y, _), (z, _), qtyp) ->
	(output_string file "I_ANTI_ADDR_INDEX ";
         output_string file x;
	 output_string file ":=";
	 output_string file y;
	 output_string file "-";
	 output_string file z;
	 output_string file "\t; ";
	 output_string file y;
	 output_string file ",";
	 output_string file z;
	 output_string file " : ";
	 pr_qtype file qtyp;
	 output_string file " *")
      | I_ADDR_SEL ((x, _), (y, _), f, offset, qtypy, qtypx) ->
	(output_string file "I_ADDR_SEL ";
         output_string file x;
	 output_string file ":=&";
	 output_string file y;
	 output_string file "->";
	 output_string file f;
	 output_string file "\t[offset=";
	 output_int file offset;
	 output_string file "]; ";
	 output_string file x;
	 output_string file " : ";
	 pr_qtype file qtypx;
	 output_string file y;
	 output_string file " : ";
	 pr_qtype file qtypy)
      | I_CALL ((x, _), f, args, qtyp) ->
	(output_string file "I_CALL ";
         output_string file x;
	 output_string file ":=";
	 output_string file f;
	 pr_args file args;
	 output_string file "\t; ";
	 output_string file x;
	 output_string file " : ";
	 pr_qtype file qtyp)
      | I_CALL_IND ((x, _), (y, _), _, args, qtyp) ->
	(output_string file "I_CALL_IND ";
         output_string file x;
	 output_string file ":=( *";
	 output_string file y;
	 output_string file ")";
	 pr_args file args;
	 output_string file "\t; ";
	 output_string file x;
	 output_string file " : ";
	 pr_qtype file qtyp) 
      | I_CONVERT ((x, _), qtyp1, qtyp2, (y, _)) ->
	(output_string file "I_CONVERT ";
         output_string file x;
	 output_string file ":=(";
	 pr_qtype file qtyp1;
	 output_string file ")";
	 output_string file y;
	 output_string file " [conversion]\t; ";
	 output_string file y;
	 output_string file " : ";
	 pr_qtype file qtyp2)
      | I_CAST ((x, _), qtyp1, qtyp2, (y, _)) ->
	(output_string file "I_CAST ";
         output_string file x;
	 output_string file ":=(";
	 pr_qtype file qtyp1;
	 output_string file ")";
	 output_string file y;
	 output_string file " [cast]\t; ";
	 output_string file y;
	 output_string file " : ";
	 pr_qtype file qtyp2)
      | I_BNOT ((x, _), (y, _), qtyp) ->
	(output_string file "I_BNOT ";
         output_string file x;
	 output_string file ":=~";
	 output_string file y;
	 output_string file "\t: ";
	 pr_qtype file qtyp)
      | I_MINUS ((x, _), (y, _), qtyp) ->
	(output_string file "I_MINUS ";
         output_string file x;
	 output_string file ":=-";
	 output_string file y;
	 output_string file "\t: ";
	 pr_qtype file qtyp)
      | I_MUL ((x, _), (y, _), (z, _), qtyp) ->
	(output_string file "I_MUL ";
         output_string file x;
	 output_string file ":=";
	 output_string file y;
	 output_string file "*";
	 output_string file z;
	 output_string file "\t: ";
	 pr_qtype file qtyp)
      | I_DIV ((x, _), (y, _), (z, _), qtyp) ->
	(output_string file "I_DIV ";
         output_string file x;
	 output_string file ":=";
	 output_string file y;
	 output_string file "/";
	 output_string file z;
	 output_string file "\t: ";
	 pr_qtype file qtyp)
      | I_MOD ((x, _), (y, _), (z, _), qtyp) ->
	(output_string file "I_MOD ";
         output_string file x;
	 output_string file ":=";
	 output_string file y;
	 output_string file "%";
	 output_string file z;
	 output_string file "\t: ";
	 pr_qtype file qtyp)
      | I_ADD ((x, _), (y, _), (z, _), qtyp) ->
	(output_string file "I_ADD ";
         output_string file x;
	 output_string file ":=";
	 output_string file y;
	 output_string file "+";
	 output_string file z;
	 output_string file "\t: ";
	 pr_qtype file qtyp)
      | I_LEFT ((x, _), (y, _), (z, _), qtyp) ->
	(output_string file "I_LEFT ";
         output_string file x;
	 output_string file ":=";
	 output_string file y;
	 output_string file "<<";
	 output_string file z;
	 output_string file "\t: ";
	 pr_qtype file qtyp)
      | I_RIGHT ((x, _), (y, _), (z, _), qtyp) ->
	(output_string file "I_RIGHT ";
         output_string file x;
	 output_string file ":=";
	 output_string file y;
	 output_string file ">>";
	 output_string file z;
	 output_string file "\t: ";
	 pr_qtype file qtyp)
      | I_AND ((x, _), (y, _), (z, _), qtyp) ->
	(output_string file "I_AND ";
         output_string file x;
	 output_string file ":=";
	 output_string file y;
	 output_string file "&";
	 output_string file z;
	 output_string file "\t: ";
	 pr_qtype file qtyp)
      | I_XOR ((x, _), (y, _), (z, _), qtyp) ->
	(output_string file "I_XOR ";
         output_string file x;
	 output_string file ":=";
	 output_string file y;
	 output_string file "^";
	 output_string file z;
	 output_string file "\t: ";
	 pr_qtype file qtyp)
      | I_OR ((x, _), (y, _), (z, _), qtyp) ->
	(output_string file "I_OR ";
         output_string file x;
	 output_string file ":=";
	 output_string file y;
	 output_string file "|";
	 output_string file z;
	 output_string file "\t: ";
	 pr_qtype file qtyp)
      | I_EQ ((x, _), (y, _), (z, _), qtyp) ->
	(output_string file "I_EQ ";
         output_string file x;
	 output_string file ":=";
	 output_string file y;
	 output_string file "==";
	 output_string file z;
	 output_string file "\t: ";
	 pr_qtype file qtyp)
      | I_LE ((x, _), (y, _), (z, _), qtyp) ->
	(output_string file "I_LE ";
         output_string file x;
	 output_string file ":=";
	 output_string file y;
	 output_string file "<=";
	 output_string file z;
	 output_string file "\t: ";
	 pr_qtype file qtyp)
      | I_LT ((x, _), (y, _), (z, _), qtyp) ->
	(output_string file "I_LT ";
         output_string file x;
	 output_string file ":=";
	 output_string file y;
	 output_string file "<";
	 output_string file z;
	 output_string file "\t: ";
	 pr_qtype file qtyp)
      | I_TEST_TRUE (x, _) ->
        (output_string file "I_TEST_TRUE ";
         output_string file x;
         output_string file " true?")
      | I_TEST_FALSE (x, _) ->
        (output_string file "I_TEST_FALSE ";
         output_string file x;
         output_string file " false?")
      | I_SKIP ->
        (
         output_string file "I_SKIP ";
         output_string file "skip")
           

let pr_instr_1 file src tgt instr =
    (output_int file src;
     output_string file " -> ";
     output_int file tgt;
     output_string file " : ";
     pr_instr file instr;
     output_char file '\n')


let pr_cg file cg =
    Hashtbl.iter (fun src succs ->
                  Hashtbl.iter (fun tgt (loc, instr) ->
                                pr_instr_1 file src tgt instr) succs) cg


let pr_cg_1 file cg n =
    for src = 1 to n do
        try let succs = Hashtbl.find cg src in
                Hashtbl.iter (fun tgt (loc, instr) ->
                              pr_instr_1 file src tgt instr) succs
        with Not_found -> ()
    done

type var_table = (string, cqualifier * locator * q_ctype) Hashtbl.t

type fun_table = (string, bool * locator * q_ctype
		  * node (* point d'entree dans le code *)
		  * node (* point de sortie *)
		  * control_graph (* graphe de controle, avec numeros de sommets
				   tries en ordre topologique inverse: s'il y a
				   une transition de m vers n et m>=n, alors c'est
				   un arc arriere, sur lequel il faut faire des
				   elargissements *)
		  * node (* soit N cette valeur: les sommets du graphe de controle
			  sont les entiers de 1 a N. *)
		  * (var, q_ctype) Hashtbl.t (* environnement de typage; est garanti
					      donner un type a chaque variable initialisee;
					      les variables jamais initialisees ne sont
					      pas dans cette table. *)
		  * (string, loc_expr) Hashtbl.t (* quelle expression represente
						  quelle variable auxiliaire *)
		  ) Hashtbl.t

let pr_cqualifier file cq =
    match cq with
	CQ_EXTERN -> output_string file "extern "
      | CQ_STATIC -> output_string file "static "
      | _ -> ()

let pr_var_table file vt =
    Hashtbl.iter (fun x (cq, loc, qtyp) ->
		      (pr_cqualifier file cq;
		       pr_type_name file qtyp x;
		       output_string file ";\n")) vt

let pr_fun_table file ft =
    Hashtbl.iter (fun x (stat, loc, qtyp, entryp, exitp, cg, n, env, abbrvs) ->
		      (if stat then output_string file "static " else ();
		       pr_type_name file qtyp x;
		       output_string file "\nstart: ";
		       output_int file entryp;
		       output_string file "; exit: ";
		       output_int file exitp;
		       output_string file "; {\n";
		       pr_cg_1 file cg n;
		       output_string file "}\n")) ft

let rec cg_of_program_1 (vt:var_table) (ft:fun_table) (decls:var_declaration list) =
    (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "cg_of_program_1 : main" in

    match decls with
	[] -> ()
      | CDECL (cq, (loc, x, qtyp))::l -> (
                                          (* DEBUG *) let _ = Debug.pRINT_DEBUG2 "cg_of_program_1 : appel de CDECL" in
                                          Hashtbl.add vt x (cq, loc, qtyp);
					  cg_of_program_1 vt ft l)
      | CFUN (stat, (loc, f, qtyp), code)::l ->
          begin
(* DEBUG *)  let _ = Debug.pRINT_DEBUG2 "cg_of_program_1 : appel de CFUN..." in
	    match qtyp with
	      (_, TFUN (T_PROTOTYPE (params, _), _)) ->

                (** FABRICE : le nom est corrig si la fonction est dclare en static. **)

                let fname = (if stat then (Ctype.static_name f) else f) in

	        let entryp, returnp, cg, abbrvs = cg_of_code fname code in
	        let rto = cg_rev_top_order cg entryp in
	        let rank, minn = get_ranking rto in
	        let cg' = cg_renumber cg rank minn in

                (** FABRICE : ICI, Bug dans les definitions de variables si on utilise cg'. **)
                (** Par contre, est ce que cela introduit d'autre bug ??                    **)

	        let env = cg_type_env cg rto params in

		(Hashtbl.add ft fname (stat, loc, qtyp,
                                   rank_number rank minn entryp,
				   rank_number rank minn returnp,
				   cg', -minn, env, abbrvs);
                 
		 cg_of_program_1 vt ft l)
            | _ -> failwith ("cg_of_program_1: function " ^ f ^ " has non-functional type")
          end

let cgs_of_program decls =
    let vt = Hashtbl.create 128 in
    let ft = Hashtbl.create 256 in
	(cg_of_program_1 vt ft decls;
	 vt, ft)

let pr_cgs file (vt, ft) =
    (pr_var_table file vt;
     pr_fun_table file ft)

