
(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *		use is subject to a specific license, obtainable from LSV.
 *)

open Cparse
open Error

type var = string * bool (* son nom, et un booleen [true si globale, false si locale] *)

type instr =
    I_SET_VAR of var * var * q_ctype (* x:=y : type;
	le type est un type entier, flottant ou pointeur,
	jamais fonction, tableau, struct, union ou void. *)
  | I_SET_ARRAY of var * var * q_ctype (* x:=a : type, ou a est un nom de tableau,
					et type est un TARRAY (...). *)
  | I_SET_FUN of var * string * q_ctype (* x:=f : type, ou f est un nom de fonction,
					 et type est un TFUN (...). *)
  | I_SET_CST of var * constant_val * q_ctype (* x:=constante : type *)
  | I_COPY of var * var * q_ctype (* *x=*y : type (copie de struct ou d'union) *)
  | I_PTR of var * var * q_ctype * q_ctype
  (* I_PTR (x, y, type-de-y, type-de-x): effectue x:=*y;
   le type-de-y est pointeur sur le type-de-x, qui est un type entier, flottant
   ou pointeur lui-meme. *)
  | I_SET_PTR of var * var * q_ctype * q_ctype
  (* I_SET_PTR (x, y, type-de-x, type-de-y): effectue *x:=y;
   le type-de-x est pointeur sur le type-de-y, qui est un type entier, flottant
   ou pointeur lui-meme. *)
  | I_ADDR_VAR of var * var * q_ctype
  (* I_ADDR_VAR (x, y, t): effectue x := &y;
   y est de type t, et x est de type (t * ). *)
  | I_ADDR_INDEX of var * var * var * q_ctype
  (* I_ADDR_INDEX (x, y, z, t): effectue x:=&y[z];
   y est de type pointeur vers t, ainsi que x;
   z est toujours de type long, possiblement const ou volatile. *)
  | I_ANTI_ADDR_INDEX of var * var * var * q_ctype
  (* I_ANTI_ADDR_INDEX (x, y, z, t): effecture x:=y-z;
   y et z sont de meme type (t * ), x est de type long. *)
  | I_ADDR_SEL of var * var * string * int * q_ctype * q_ctype
  (* I_ADDR_SEL (x, y, f, offset, type-de-y, type-de-x): effectue x:=&y->f,
   ou offset est l'offset du champ f par rapport au debut de la structure y. *)
  | I_CALL of var * string * (var * q_ctype) list * q_ctype (* x:=f(x1,...,xn) *)
  | I_CALL_IND of var * var * q_ctype * (var * q_ctype) list * q_ctype
  (* I_CALL_IND (x, y, type-de-y, [x1;...;xn], type-retour)
   effectue x:= ( *y)(x1,...,xn) : type-retour *)
  | I_CONVERT of var * q_ctype * q_ctype * var
  (* I_CONVERT (x, T1, T2, y): x:=(T1)y avec y:T2;
   il s'agit d'une conversion (int -> double, etc.), qui implique un changement de
   representation, contrairement au cast, qui est juste un changement de type.
   On a uniquement les cas suivants:
   float <-> double
   long <-> double
   long <-> unsigned long
   int <-> unsigned int
   short <-> unsigned short
   char <-> unsigned char
   long <-> int
   unsigned long <-> unsigned int
   long <-> short
   unsigned long <-> unsigned short
   long <-> char
   unsigned long <-> unsigned char
   int <-> short
   unsigned int <-> unsigned short
   int <-> char
   unsigned int <-> unsigned char
   short <-> char
   unsigned short <-> unsigned char
   unsigned long <-> T * (pour n'importe quel type T)
   T [...] -> T * (on peut convertir un type tableau en un type pointeur correspondant,
                   mais pas le contraire)
   *)
  | I_CAST of var * q_ctype * q_ctype * var (* x:=(T1)y avec y:T2 *)
  | I_BNOT of var * var * q_ctype (* x:=~y : type *)
  (* Le type est char, short, int, ou long, possiblement unsigned,
   possiblement const ou volatile.  On notera que
   ~y = -y-1 modulo les bornes de definition du type entier considere. *)
  | I_MINUS of var * var * q_ctype (* x:=-y : type (type est un type numerique,
				    int, char, short, long
				    [possiblement unsigned], float, double;
				    possiblement const ou volatile) *)
  | I_MUL of var * var * var * q_ctype (* x:=y*z *)
    (* multiplication, le type est char, short, int, long, float
     ou double, possiblement unsigned, const ou volatile. *)
  | I_DIV of var * var * var * q_ctype (* x:=y/z *)
    (* division, le type est char, short, int, long, float
     ou double, possiblement unsigned, const ou volatile. *)
  | I_MOD of var * var * var * q_ctype (* x:=y%z *)
    (* reste, le type est char, short, int, ou long,
     possiblement unsigned, const ou volatile. *)
  | I_ADD of var * var * var * q_ctype (* x:=y+z *)
    (* addition, le type est char, short, int, long, float
     ou double, possiblement unsigned, const ou volatile. *)
  | I_LEFT of var * var * var * q_ctype (* x:=y<<z *)
    (* le type est celui de x et de y; c'est char, short, int, long, float
     ou double, possiblement unsigned, const ou volatile.
     Le type de z est toujours int. *)
  | I_RIGHT of var * var * var * q_ctype (* x:=y>>z *)
    (* le type est celui de x et de y; c'est char, short, int, long, float
     ou double, possiblement unsigned, const ou volatile.
     Le type de z est toujours int. *)
  | I_AND of var * var * var * q_ctype (* x:=y&z *)
    (* le type est int, unsigned int, long ou unsigned long. *)
  | I_XOR of var * var * var * q_ctype (* x:=y^z *)
    (* le type est char, short, int, long, possiblement unsigned, const ou volatile. *)
  | I_OR of var * var * var * q_ctype (* x:=y|z *)
    (* le type est char, short, int, long, possiblement unsigned, const ou volatile. *)
  | I_EQ of var * var * var * q_ctype (* x:=(y==z) avec y,z:type *)
    (* le type est char, short, int, long, possiblement unsigned,
     ou float ou double, possiblement const ou volatile,
     ou bien un type non numerique (pointeur, tableau).
     Le type de x est toujours int. *)
  | I_LE of var * var * var * q_ctype (* x:=(y<=z) avec y,z:type *)
    (* le type en question est char, short, int, long, float
     ou double, possiblement unsigned, const ou volatile.
     Le type de x est toujours int. *)
  | I_LT of var * var * var * q_ctype (* x:=(y<z) avec y,z:type *)
    (* le type en question est char, short, int, long, float
     ou double, possiblement unsigned, const ou volatile.
     Le type de x est toujours int. *)
  | I_TEST_TRUE of var (* transition prise si x est vraie (entier non nul) *)
  | I_TEST_FALSE of var (* transition prise si x est fausse (entier nul) *)
  | I_SKIP (* ne fait rien. *)

type loc_instr = locator * instr

type node = int

type control_graph = (node, (node, loc_instr) Hashtbl.t) Hashtbl.t
    (* envoie chaque sommet [un int] vers une map
     qui envoie chaque successeur [un int encore]
     vers une instruction.
     Il s'agit du graphe de controle d'une seule
     fonction. *)

(* Un point de programme local: le nom de la fonction
 dans laquelle on est, plus le sommet [un int] dans le
 graphe de controle associe. *)
type loc_ppoint = string * node

(* Un (vrai) point de programme =
 un point de programme local (la ou on est),
 plus un autre point de programme local (d'ou on a ete appele)
 ou None si on est dans main(). *)
type ppoint = loc_ppoint * loc_ppoint option

type var_table = (string, cqualifier * locator * q_ctype) Hashtbl.t
type fun_table = (string, bool * locator * q_ctype
		  * node (* point d'entree dans le code *)
		  * node (* point de sortie *)
		  * control_graph (* graphe de controle, avec numeros de sommets
				   tries en ordre topologique inverse: s'il y a
				   une transition de m vers n et m>=n, alors c'est
				   un arc arriere, sur lequel il faut faire des
				   elargissements *)
		  * node (* soit N cette valeur: les sommets du graphe de controle
			  sont les entiers de 1 a N. *)
		  * (var, q_ctype) Hashtbl.t (* environnement de typage; est garanti
					      donner un type a chaque variable initialisee;
					      les variables jamais initialisees ne sont
					      pas dans cette table. *)
		  * (string, loc_expr) Hashtbl.t (* quelle expression represente
						  quelle variable auxiliaire *)
		  ) Hashtbl.t

val cgs_of_program : var_declaration list -> var_table * fun_table

val pr_cgs : out_channel -> var_table * fun_table -> unit
val pr_instr_1 : out_channel -> node -> node -> instr -> unit
