
(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *		use is subject to a specific license, obtainable from LSV.
 *)

open Const
open Arith
open Cparse
open Sigs
open Control
open Intervalle
open Itv_int
open Itv_float
open Zones
open Glue

let prerr_itv itv =
    match itv with
	ITV (a, b) -> if a=b
			  then prerr_float a
		      else (prerr_string "[";
			    prerr_float a;
			    prerr_string ",";
			    prerr_float b;
			    prerr_string "]")
      | ITV_EMPTY -> prerr_string "empty"

let offset_zero = ITV (0.0, 0.0)

let prerr_offset sep offset =
     if itv_equal offset offset_zero
	 then ()
     else (prerr_string sep;
	   prerr_itv offset)

let output_loc_pp file ft (f, node) =
    output_string file f;
    try let (st, loc, rest, entryp, exitp, cg, n, typenv, abbrvs) = Hashtbl.find ft f in
        let succs = Hashtbl.find cg node in
	    Hashtbl.iter (fun tgt ((fl, line, _, _, _), _) ->
			      (output_string file ", file \"";
			       output_string file fl;
			       output_string file "\", line ";
			       output_string file (string_of_int line);
			       raise Not_found)
			      ) succs
    with Not_found -> ()

let output_pp file ft (locpp, ctx) =
    (output_loc_pp file ft locpp;
     match ctx with
	 None -> ()
       | Some call -> (output_string file " called from ";
		       output_loc_pp file ft call))

let prerr_pp ft pp = output_pp stderr ft pp

let prerr_zone ft z =
    match z with
	Z_NULL -> prerr_string "NULL"
      | Z_GLOBAL (x, itv) -> (prerr_string x;
			      prerr_offset "#" itv)
      | Z_HEAP pp -> prerr_pp ft pp
      | Z_STACK (x, itv, f, ctx) ->
	(prerr_string x;
	 prerr_string " (";
	 prerr_string f;
	 (match ctx with
	      None -> ()
	    | Some locpp ->
	      (prerr_string ", ";
	       output_loc_pp stderr ft locpp));
	 prerr_string ")";
	 prerr_offset "#" itv)
      | Z_STRING s -> (prerr_string "\"";
		       prerr_string (String.escaped s);
		       prerr_string "\"")
      | Z_FUN f -> prerr_string f
	 

let prerr_zone_table ft zt =
    let sep = ref "" in
	ZoneTable.iter (fun z offset ->
			    (prerr_string (!sep);
			     sep := "|";
			     prerr_zone ft z;
			     prerr_offset "+" offset
			     )
			    ) zt

let prerr_abs_addr ft a =
    match a with
	None -> prerr_string "*hyperspace*"
      | Some zt -> prerr_zone_table ft zt

let prerr_val ft v =
    match v with
	ITV_INT itv -> prerr_itv itv
      | ITV_FLOAT (itv, nan) -> (prerr_itv itv;
				 if nan
				     then prerr_string " U {NaNs}")
      | ABS_ADDR ad -> prerr_abs_addr ft ad

let ae_print_diff_1 ft ae1 ae2 =
    (AbsEnv.iter (fun z i2 ->
		      try let i1 = AbsEnv.find z ae1 in
		              if av_equal i1 i2
				  then ()
			      else (prerr_string " ";
				    prerr_string z;
				    prerr_string ":=";
				    prerr_val ft i2;
				    prerr_string ";")
		      with Not_found ->
			   (prerr_string " ";
			    prerr_string z;
			    prerr_string ":=";
			    prerr_val ft i2;
			    prerr_string ";")
		      ) ae2;
     AbsEnv.iter (fun z i1 ->
		      try ignore (AbsEnv.find z ae2)
		      with Not_found ->
			   (prerr_string " ";
			    prerr_string z;
			    prerr_string ":=top;")
		      ) ae1
     )

let ae_print_diff ft (ae_glob1, ae_loc1) (ae_glob2, ae_loc2) =
    (ae_print_diff_1 ft ae_loc1 ae_loc2;
     ae_print_diff_1 ft ae_glob1 ae_glob2)
